"""
Server for Commune modules.
"""

import json
import random
import re
from datetime import datetime, timezone
from functools import partial
from typing import Any, Awaitable, Callable

import fastapi
import starlette.datastructures
from fastapi import APIRouter, Request, Response
from fastapi.responses import JSONResponse
from fastapi.routing import APIRoute
from pydantic import BaseModel
from scalecodec.utils.ss58 import ss58_encode  # type: ignore
from substrateinterface import Keypair  # type: ignore

from communex._common import get_node_url
from communex.client import CommuneClient
from communex.key import check_ss58_address
from communex.module import _signer as signer
from communex.module._rate_limiters.limiters import (
    IpLimiterMiddleware, StakeLimiterMiddleware, 
    StakeLimiterParams, IpLimiterParams
    )

from communex.module.module import Module, endpoint, EndpointDefinition
from communex.module._util import log
from communex.types import Ss58Address
from communex.util.memo import TTLDict

# Regular expression to match a hexadecimal number
HEX_PATTERN = re.compile(r"^[0-9a-fA-F]+$")


# TODO: merge `is_hex_string` into `parse_hex`
def is_hex_string(string: str):
    return bool(HEX_PATTERN.match(string))


def parse_hex(hex_str: str) -> bytes:
    if hex_str[0:2] == "0x":
        return bytes.fromhex(hex_str[2:])
    else:
        return bytes.fromhex(hex_str)


def build_input_handler_route_class(
    subnets_whitelist: list[int] | None,
    module_key: Ss58Address,
    request_staleness: int,
    blockchain_cache: TTLDict[str, list[Ss58Address]],
    host_key: Keypair,
    use_testnet: bool,
) -> type[APIRoute]:
    class InputHandlerRoute(APIRoute):
        def get_route_handler(self):
            original_route_handler = super().get_route_handler()

            async def custom_route_handler(request: Request) -> Response:
                body = await request.body()

                # TODO: we'll replace this by a Result ADT :)
                match self._check_inputs(request, body, module_key):
                    case (False, error):
                        return error
                    case (True, _):
                        pass
               
                body_dict: dict[str, dict[str, Any]] = json.loads(body)
                timestamp = body_dict['params'].get("timestamp", None)
                legacy_timestamp = request.headers.get("X-Timestamp", None)
                try:
                    timestamp_to_use = timestamp if not legacy_timestamp else legacy_timestamp
                    request_time = datetime.fromisoformat(timestamp_to_use)
                except Exception:
                    return JSONResponse(status_code=400, content={"error": "Invalid ISO timestamp given"})
                if (datetime.now(timezone.utc) - request_time).total_seconds() > request_staleness:
                    return JSONResponse(status_code=400, content={"error": "Request is too stale"})
                response: Response = await original_route_handler(request)
                return response

            return custom_route_handler


        @staticmethod
        def _check_inputs(request: Request, body: bytes, module_key: Ss58Address):
            required_headers = ["x-signature", "x-key", "x-crypto"]
            optional_headers = ["x-timestamp"]

            # TODO: we'll replace this by a Result ADT :)
            match _get_headers_dict(request.headers, required_headers, optional_headers):
                case (False, error):
                    return (False, error)
                case (True, headers_dict):
                    pass

            # TODO: we'll replace this by a Result ADT :)
            match _check_signature(headers_dict, body, module_key):
                case (False, error):
                    return (False, error)
                case (True, _):
                    pass

            # TODO: we'll replace this by a Result ADT :)
            match _check_key_registered(
                subnets_whitelist,
                headers_dict,
                blockchain_cache,
                host_key,
                use_testnet,
            ):
                case (False, error):
                    return (False, error)
                case (True, _):
                    pass

            return (True, None)

    return InputHandlerRoute


def _json_error(code: int, message: str):
    return JSONResponse(status_code=code, content={"error": {"code": code, "message": message}})


def _get_headers_dict(
    headers: starlette.datastructures.Headers,
    required: list[str],
    optional: list[str],
):
    headers_dict: dict[str, str] = {}
    for required_header in required:
        value = headers.get(required_header)
        if not value:
            code = 400
            return False, _json_error(code, f"Missing header: {required_header}")
        headers_dict[required_header] = value
    for optional_header in optional:
        value = headers.get(optional_header)
        if value:
            headers_dict[optional_header] = value

    return True, headers_dict


# TODO: type `headers_dict` better
def _check_signature(
    headers_dict: dict[str, str],
    body: bytes,
    module_key: Ss58Address
):
    key = headers_dict["x-key"]
    signature = headers_dict["x-signature"]
    crypto = int(headers_dict["x-crypto"])  # TODO: better handling of this

    if not is_hex_string(key):
        return (False, _json_error(400, "X-Key should be a hex value"))

    signature = parse_hex(signature)
    key = parse_hex(key)

    timestamp = headers_dict.get("x-timestamp")
    legacy_verified = False
    if timestamp:
        # tries to do a legacy verification
        json_body = json.loads(body)
        json_body["timestamp"] = timestamp
        stamped_body = json.dumps(json_body).encode()
        legacy_verified = signer.verify(key, crypto, stamped_body, signature)

    verified = signer.verify(key, crypto, body, signature)
    if not verified and not legacy_verified:
        return (False, _json_error(401, "Signatures doesn't match"))

    body_dict: dict[str, dict[str, Any]] = json.loads(body)
    target_key = body_dict['params'].get("target_key", None)
    if not target_key or target_key != module_key:
        return (False, _json_error(401, "Wrong target_key in body"))

    return (True, None)


def _make_client(node_url: str):
    return CommuneClient(url=node_url, num_connections=1, wait_for_finalization=False)


def _check_key_registered(
        subnets_whitelist: list[int] | None,
        headers_dict: dict[str, str],
        blockchain_cache: TTLDict[str, list[Ss58Address]],
        host_key: Keypair,
        use_testnet: bool,
):
    key = headers_dict["x-key"]
    if not is_hex_string(key):
        return (False, _json_error(400, "X-Key should be a hex value"))
    key = parse_hex(key)

    # TODO: checking for key being registered should be smarter
    # e.g. query and store all registered modules periodically.
    node_url = get_node_url(None, use_testnet=use_testnet)
    # TODO: client pool for entire module server
    client = _make_client(node_url)  # TODO: get client from outer context
    ss58_format = 42
    ss58 = ss58_encode(key, ss58_format)
    ss58 = check_ss58_address(ss58, ss58_format)

    # If subnets whitelist is specified, checks if key is registered in one
    # of the given subnets
    def query_keys(subnet: int):
        return [*client.query_map_key(subnet).values()]
    allowed_subnets: dict[int, bool] = {}
    caller_subnets: list[int] = []
    if subnets_whitelist is not None:
        for subnet in subnets_whitelist:
            get_keys_on_subnet = partial(query_keys, subnet)
            cache_key = f"keys_on_subnet_{subnet}"
            keys_on_subnet = blockchain_cache.get_or_insert_lazy(
                cache_key, get_keys_on_subnet
            )
            if host_key.ss58_address not in keys_on_subnet:
                log(
                    f"WARNING: This miner is deregistered on subnet {subnet}"
                )
            else:
                allowed_subnets[subnet] = True
            if ss58 in keys_on_subnet:
                caller_subnets.append(subnet)
        
        if not allowed_subnets:
            log("WARNING: Miner is not registered on any subnet")
            return False, _json_error(403, "Miner is not registered on any subnet")
        
        # searches for a common subnet between caller and miner
        # TODO: use sets
        allowed_subnets = {
            subnet: allowed for subnet, allowed in allowed_subnets.items() if (
                subnet in caller_subnets
                )
            }
        if not allowed_subnets:
            return False, _json_error(
                403, "Caller key is not registered in any subnet that the miner is"
            )
    else:
        # accepts everything
        pass
    
    return (True, None)


Callback = Callable[[Request], Awaitable[Response]]


class ModuleServer:
    def __init__(
        self,
        module: Module,
        key: Keypair,
        max_request_staleness: int = 120,
        whitelist: list[str] | None = None,
        blacklist: list[str] | None = None,
        subnets_whitelist: list[int] | None = None,
        lower_ttl: int = 600,
        upper_ttl: int = 700,
        limiter: StakeLimiterParams | IpLimiterParams = StakeLimiterParams(),
        ip_blacklist: list[str] | None = None,
        use_testnet: bool = False,
    ) -> None:
        self._module = module
        self._app = fastapi.FastAPI()
        self._subnets_whitelist = subnets_whitelist
        self.key = key
        self.max_request_staleness = max_request_staleness
        self._blacklist = blacklist
        self._whitelist = whitelist
        ttl = random.randint(lower_ttl, upper_ttl)
        self._blockchain_cache = TTLDict[str, list[Ss58Address]](ttl)
        self._ip_blacklist = ip_blacklist


        # Midlewares

        if isinstance(limiter, StakeLimiterParams):
            self._app.add_middleware(
                StakeLimiterMiddleware, 
                subnets_whitelist=self._subnets_whitelist,
                params=limiter,
                )
        else:
            self._app.add_middleware(
                IpLimiterMiddleware, params=limiter
                )

        self.register_extra_middleware()

        # Routes
        self._router = APIRouter(
            route_class=build_input_handler_route_class(
                self._subnets_whitelist,
                check_ss58_address(self.key.ss58_address),
                self.max_request_staleness,
                self._blockchain_cache,
                self.key,
                use_testnet
            )
        )
        self.register_endpoints(self._router)
        self._app.include_router(self._router)


    def get_fastapi_app(self):
        return self._app


    def register_endpoints(self, router: APIRouter):
        endpoints = self._module.get_endpoints()
        for name, endpoint_def in endpoints.items():

            class Body(BaseModel):
                params: endpoint_def.params_model  # type: ignore

            def handler(end_def: EndpointDefinition[Any, ...], body: Body):

                return end_def.fn(self._module, **body.params.model_dump())  # type: ignore

            defined_handler = partial(handler, endpoint_def)
            router.post(f"/method/{name}")(defined_handler)


    def register_extra_middleware(self):
        async def check_lists(request: Request, call_next: Callback):

            if not request.url.path.startswith('/method'):
                return await call_next(request)
            key = request.headers.get("x-key")
            if not key:
                return _json_error(400, "Missing header: X-Key")
            ss58_format = 42
            ss58 = ss58_encode(key, ss58_format)
            ss58 = check_ss58_address(ss58, ss58_format)

            if request.client is None:
                return _json_error(400, "Address should be present in request")
            if self._blacklist and ss58 in self._blacklist:
                return _json_error(403, "You are blacklisted")
            if self._ip_blacklist and request.client.host in self._ip_blacklist:
                return _json_error(403, "Your IP is blacklisted")
            if self._whitelist and ss58 not in self._whitelist:
                return _json_error(403, "You are not whitelisted")
            response = await call_next(request)
            return response

        self._app.middleware("http")(check_lists)


    def add_to_blacklist(self, ss58_address: str | Ss58Address):
        if not self._blacklist:
            self._blacklist = []
        self._blacklist.append(ss58_address)


    def add_to_whitelist(self, ss58_address: str | Ss58Address):
        if not self._whitelist:
            self._whitelist = []
        self._whitelist.append(ss58_address)


def main():
    class Amod(Module):
        @endpoint
        def do_the_thing(self, awesomness: int = 42):
            if awesomness > 60:
                msg = f"You're super awesome: {awesomness} awesomness"
            else:
                msg = f"You're not that awesome: {awesomness} awesomness"
            return {"msg": msg}

    a_mod = Amod()
    keypair = Keypair.create_from_mnemonic(signer.TESTING_MNEMONIC)
    server = ModuleServer(a_mod, keypair, subnets_whitelist=None)
    app = server.get_fastapi_app()

    import uvicorn
    uvicorn.run(app, host="127.0.0.1", port=8000)  # type: ignore


if __name__ == "__main__":
    main()



import datetime

from typing import Any, Literal


def iso_timestamp_now() -> str:
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    iso_now = now.isoformat()
    return iso_now


def log(
        msg: str,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: Any | None = None,
        flush: Literal[False] = False
    ):
    print(f"[{iso_timestamp_now()}] " + msg, *values, sep=sep, end=end, file=file, flush=flush)


from typing import Awaitable, Callable

from fastapi import Request, Response
from fastapi.responses import JSONResponse
from keylimiter import TokenBucketLimiter
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.types import ASGIApp
from pydantic_settings import BaseSettings
from ._stake_limiter import StakeLimiter

Callback = Callable[[Request], Awaitable[Response]]


class IpLimiterParams(BaseSettings):
    bucket_size: int = 15
    refill_rate: int = 1

    class config:
        env_prefix = "CONFIG_IP_LIMITER_"
        extra = "ignore"

class StakeLimiterParams(BaseSettings):
    epoch: int = 800
    cache_age: int = 600
    get_refill_per_epoch: Callable[[int], float] | None = None
    token_ratio: int = 1

    class config:
        env_prefix = "CONFIG_STAKE_LIMITER_"
        extra = "ignore"


class IpLimiterMiddleware(BaseHTTPMiddleware):
    def __init__(
            self,
            app: ASGIApp,
            params: IpLimiterParams | None,
    ):
        '''
        :param app: FastAPI instance
        :param limiter: KeyLimiter instance OR None

        If limiter is None, then a default TokenBucketLimiter is used with the following config:
        bucket_size=200, refill_rate=15
        '''
        super().__init__(app)

        # fallback to default limiter
        if not params:
            params = IpLimiterParams()
        self._limiter = TokenBucketLimiter(bucket_size=params.bucket_size, refill_rate=params.refill_rate)

    async def dispatch(self, request: Request, call_next: Callback) -> Response:
        assert request.client is not None, "request is invalid"
        assert request.client.host, "request is invalid."

        ip = request.client.host

        is_allowed = self._limiter.allow(ip)

        if not is_allowed:
            response = JSONResponse(
                status_code=429, 
                headers={"X-RateLimit-Remaining": str(self._limiter.remaining(ip))},
                content={"error": "Rate limit exceeded"}
                )
            return response

        response = await call_next(request)

        return response


class StakeLimiterMiddleware(BaseHTTPMiddleware):
    def __init__(
            self,
            app: ASGIApp,
            subnets_whitelist: list[int] | None = [0],
            params: StakeLimiterParams | None = None,
    ):

        super().__init__(app)

        if not params:
            params = StakeLimiterParams()
        self._limiter = StakeLimiter(
            subnets_whitelist, 
            epoch=params.epoch, 
            max_cache_age=params.cache_age,
            get_refill_rate=params.get_refill_per_epoch,
            )

    async def dispatch(self, request: Request, call_next: Callback) -> Response:
        if request.client is None:
            response = JSONResponse(
                status_code=401,
                content={
                    "error": "Address should be present in request"
                }
            )
            return response
        key = request.headers.get('x-key')
        if not key:
            response = JSONResponse(
                status_code=401,
                content={"error": "Valid X-Key not provided on headers"}
                )
            return response
        

        is_allowed = await self._limiter.allow(key)

        if not is_allowed:
            response = JSONResponse(
                status_code=429, 
                headers={"X-RateLimit-TryAfter": f"{str(await self._limiter.retry_after(key))} seconds"},
                content={"error": "Rate limit exceeded"}
                )
            return response
        response = await call_next(request)

        return response



"""
Common types for the communex module.
"""

from typing import NewType, TypedDict

Ss58Address = NewType("Ss58Address", str)
"""Substrate SS58 address.

The `SS58 encoded address format`_ is based on the Bitcoin Base58Check format,
but with a few modification specifically designed to suite Substrate-based
chains.

.. _SS58 encoded address format:
    https://docs.substrate.io/reference/address-formats/
"""


# TODO: replace with dataclasses


class NetworkParams(TypedDict):
    max_allowed_modules: int
    max_registrations_per_block: int
    target_registrations_interval: int  #  in blocks
    target_registrations_per_interval: int
    unit_emission: int
    max_name_length: int
    min_name_length: int
    burn_rate: int
    min_burn: int  # min burn to register
    max_burn: int  # max burn to register
    min_stake: int
    min_weight_stake: int
    max_allowed_subnets: int
    adjustment_alpha: int
    floor_delegation_fee: int
    max_allowed_weights: int
    curator: Ss58Address
    proposal_cost: int
    proposal_expiration: int
    proposal_participation_threshold: int
    subnet_stake_threshold: int


class SubnetParams(TypedDict):
    founder: Ss58Address
    founder_share: int
    immunity_period: int
    incentive_ratio: int
    max_allowed_uids: int
    max_allowed_weights: int
    min_allowed_weights: int
    max_stake: int
    max_weight_age: int
    min_stake: int
    name: str
    tempo: int
    trust_ratio: int
    vote_mode: str
    bonds_ma: int | None
    maximum_set_weight_calls_per_epoch: int | None


# redundant "TypedDict" inheritance because of pdoc warns.
# see https://github.com/mitmproxy/pdoc/blob/26d40827ddbe1658e8ac46cd092f17a44cf0287b/pdoc/doc.py#L691-L692
class SubnetParamsWithEmission(SubnetParams, TypedDict):
    """SubnetParams with emission field.
    """
    emission: int
    """Subnet emission percentage (0-100).
    """


class ModuleInfo(TypedDict):
    uid: int
    key: Ss58Address
    name: str
    address: str  # "<ip>:<port>"
    emission: int
    incentive: int
    dividends: int
    stake_from: list[tuple[str, int]]  # TODO: type key with Ss58Address
    regblock: int  # block number
    last_update: int  # block number
    stake: int
    delegation_fee: int
    metadata: str


class ModuleInfoWithBalance(ModuleInfo):
    balance: int


class ModuleInfoWithOptionalBalance(ModuleInfo):
    balance: int | None



import re
from typing import Any

from communex.client import CommuneClient
from communex.key import check_ss58_address
from communex.types import (ModuleInfoWithOptionalBalance, NetworkParams,
                            Ss58Address, SubnetParamsWithEmission)

IPFS_REGEX = re.compile(r'^Qm[1-9A-HJ-NP-Za-km-z]{44}$')


def get_map_modules(
    client: CommuneClient,
    netuid: int = 0,
    include_balances: bool = False,
) -> dict[str, ModuleInfoWithOptionalBalance]:
    """
    Gets all modules info on the network
    """

    request_dict: dict[Any, Any] = {
        "SubspaceModule": [
            ("StakeFrom", [netuid]),
            ("Keys", [netuid]),
            ("Name", [netuid]),
            ("Address", [netuid]),
            ("RegistrationBlock", [netuid]),
            ('DelegationFee', [netuid]),
            ('Emission', []),
            ('Incentive', []),
            ("Dividends", []),
            ('LastUpdate', []),
            ("Metadata", [netuid]),
        ],
    }
    if include_balances:
        request_dict["System"] = [("Account", [])]

    bulk_query = client.query_batch_map(request_dict)
    ss58_to_stakefrom, uid_to_key, uid_to_name, uid_to_address, uid_to_regblock, \
        ss58_to_delegationfee, uid_to_emission, uid_to_incentive, uid_to_dividend, \
        uid_to_lastupdate, ss58_to_balances, uid_to_metadata = (
            bulk_query.get("StakeFrom", {}),
            bulk_query.get("Keys", {}),
            bulk_query["Name"],
            bulk_query["Address"],
            bulk_query["RegistrationBlock"],
            bulk_query["DelegationFee"],
            bulk_query["Emission"],
            bulk_query["Incentive"],
            bulk_query["Dividends"],
            bulk_query["LastUpdate"],
            bulk_query.get("Account", {}),
            bulk_query.get("Metadata", {}),

        )

    result_modules: dict[str, ModuleInfoWithOptionalBalance] = {}

    for uid, key in uid_to_key.items():
        key = check_ss58_address(key)

        name = uid_to_name[uid]
        address = uid_to_address[uid]
        emission = uid_to_emission[netuid][uid]
        incentive = uid_to_incentive[netuid][uid]
        dividend = uid_to_dividend[netuid][uid]
        regblock = uid_to_regblock[uid]
        stake_from = ss58_to_stakefrom.get(key, [])
        last_update = uid_to_lastupdate[netuid][uid]
        delegation_fee = ss58_to_delegationfee.get(
            key, 20)  # 20% default delegation fee
        metadata = uid_to_metadata.get(uid, None)

        balance = None
        if include_balances and ss58_to_balances is not None:  # type: ignore
            balance_dict = ss58_to_balances.get(key, None)
            if balance_dict is not None:
                assert isinstance(balance_dict['data'], dict)
                balance = balance_dict['data']['free']
            else:
                balance = 0
        stake = sum(stake for _, stake in stake_from)

        module: ModuleInfoWithOptionalBalance = {
            "uid": uid,
            "key": key,
            "name": name,
            "address": address,
            "emission": emission,
            "incentive": incentive,
            "dividends": dividend,
            "stake_from": stake_from,
            "regblock": regblock,
            "last_update": last_update,
            "balance": balance,
            "stake": stake,
            "delegation_fee": delegation_fee,
            "metadata": metadata
        }

        result_modules[key] = module
    return result_modules


def get_map_subnets_params(
    client: CommuneClient,
    block_hash: str | None = None
) -> dict[int, SubnetParamsWithEmission]:
    """
    Gets all subnets info on the network
    """

    bulk_query = client.query_batch_map(
        {
            "SubspaceModule": [
                ("ImmunityPeriod", []),
                ("MinAllowedWeights", []),
                ("MaxAllowedWeights", []),
                ('MinStake', []),
                ("SubnetEmission", []),
                ('MaxStake', []),
                ("Tempo", []),
                ("MaxAllowedUids", []),
                ('Founder', []),
                ("FounderShare", []),
                ('IncentiveRatio', []),
                ('TrustRatio', []),
                ('VoteModeSubnet', []),
                ('SubnetNames', []),
                ('MaxWeightAge', []),
                ("BondsMovingAverage", []),
                ("MaximumSetWeightCallsPerEpoch", [])
            ],
        },
        block_hash
    )
    (
        netuid_to_emission, netuid_to_tempo, netuid_to_immunity_period,
        netuid_to_min_allowed_weights, netuid_to_max_allowed_weights,
        netuid_to_max_allowed_uids, netuid_to_min_stake,
        netuid_to_max_stake, netuid_to_founder, netuid_to_founder_share,
        netuid_to_incentive_ratio, netuid_to_trust_ratio,
        netuid_to_vote_mode_subnet,
        netuid_to_subnet_names,
        netuid_to_weight_age,
        netuid_to_bonds_ma,
        netuid_to_maximum_weight_calls_per_epoch
    ) = (
        bulk_query["SubnetEmission"], bulk_query["Tempo"],
        bulk_query["ImmunityPeriod"], bulk_query["MinAllowedWeights"],
        bulk_query["MaxAllowedWeights"], bulk_query["MaxAllowedUids"],
        bulk_query["MinStake"], bulk_query["MaxStake"],
        bulk_query["Founder"], bulk_query["FounderShare"],
        bulk_query["IncentiveRatio"], bulk_query["TrustRatio"],
        bulk_query["VoteModeSubnet"],
        bulk_query["SubnetNames"], bulk_query["MaxWeightAge"],
        bulk_query.get("BondsMovingAverage", {}),
        bulk_query.get("MaximumSetWeightCallsPerEpoch", {}),
    )
    result_subnets: dict[int, SubnetParamsWithEmission] = {}

    for netuid, name in netuid_to_subnet_names.items():
        name = name
        founder = Ss58Address(netuid_to_founder[netuid])
        founder_share = netuid_to_founder_share[netuid]
        immunity_period = netuid_to_immunity_period[netuid]
        incentive_ratio = netuid_to_incentive_ratio[netuid]
        max_allowed_uids = netuid_to_max_allowed_uids[netuid]
        max_allowed_weights = netuid_to_max_allowed_weights[netuid]
        min_allowed_weights = netuid_to_min_allowed_weights[netuid]
        max_stake = netuid_to_max_stake[netuid]
        min_stake = netuid_to_min_stake[netuid]
        tempo = netuid_to_tempo[netuid]
        trust_ratio = netuid_to_trust_ratio[netuid]
        vote_mode = netuid_to_vote_mode_subnet[netuid]
        emission = netuid_to_emission[netuid]
        max_weight_age = netuid_to_weight_age[netuid]
        bonds_ma = netuid_to_bonds_ma.get(netuid, None)
        maximum_weight_calls = netuid_to_maximum_weight_calls_per_epoch.get(
            netuid, None)

        subnet: SubnetParamsWithEmission = {
            "name": name,
            "founder": founder,
            "founder_share": founder_share,
            "immunity_period": immunity_period,
            "incentive_ratio": incentive_ratio,
            "max_allowed_uids": max_allowed_uids,
            "max_allowed_weights": max_allowed_weights,
            "min_allowed_weights": min_allowed_weights,
            "max_stake": max_stake,
            "min_stake": min_stake,
            "tempo": tempo,
            "trust_ratio": trust_ratio,
            "vote_mode": vote_mode,
            "emission": emission,
            "max_weight_age": max_weight_age,
            "bonds_ma": bonds_ma,
            "maximum_set_weight_calls_per_epoch": maximum_weight_calls
        }

        result_subnets[netuid] = subnet

    return result_subnets


def get_global_params(c_client: CommuneClient) -> NetworkParams:
    """
    Returns global parameters of the whole commune ecosystem
    """

    query_all = c_client.query_batch({
        "SubspaceModule": [
            ("MaxAllowedSubnets", []),
            ("MaxAllowedModules", []),
            ("MaxRegistrationsPerBlock", []),
            ("TargetRegistrationsInterval", []),
            ("TargetRegistrationsPerInterval", []),
            ("UnitEmission", []),
            ("MaxNameLength", []),
            ("BurnRate", []),
            ("MinBurn", []),
            ("MaxBurn", []),
            ("MinStakeGlobal", []),
            ("MinWeightStake", []),
            ("AdjustmentAlpha", []),
            ("FloorDelegationFee", []),
            ("MaxAllowedWeightsGlobal", []),
            ("Curator", []),
            ("ProposalCost", []),
            ("ProposalExpiration", []),
            ("ProposalParticipationThreshold", []),
            ("Curator", []),
            ("SubnetStakeThreshold", []),
            ("MinWeightStake", []),
            ("MinNameLength", []),
        ],
    })

    global_params: NetworkParams = {
        "max_allowed_subnets": int(query_all["MaxAllowedSubnets"]),
        "max_allowed_modules": int(query_all["MaxAllowedModules"]),
        "max_registrations_per_block": int(query_all["MaxRegistrationsPerBlock"]),
        "target_registrations_interval": int(query_all["TargetRegistrationsInterval"]),
        "target_registrations_per_interval": int(query_all["TargetRegistrationsPerInterval"]),
        "unit_emission": int(query_all["UnitEmission"]),
        "max_name_length": int(query_all["MaxNameLength"]),
        "burn_rate": int(query_all["BurnRate"]),
        "min_burn": int(query_all["MinBurn"]),
        "max_burn": int(query_all["MaxBurn"]),
        "min_stake": int(query_all["MinStakeGlobal"]),
        "min_weight_stake": int(query_all["MinWeightStake"]),
        "adjustment_alpha": int(query_all["AdjustmentAlpha"]),
        "floor_delegation_fee": int(query_all["FloorDelegationFee"]),
        "max_allowed_weights": int(query_all["MaxAllowedWeightsGlobal"]),
        "curator": Ss58Address(query_all["Curator"]),
        "proposal_cost": int(query_all["ProposalCost"]),
        "proposal_expiration": int(query_all["ProposalExpiration"]),
        "proposal_participation_threshold": int(query_all["ProposalParticipationThreshold"]),
        "subnet_stake_threshold": int(query_all["SubnetStakeThreshold"]),
        "min_name_length": int(query_all["MinNameLength"]),
    }

    return global_params


def concat_to_local_keys(
        balance: dict[str, int],
        local_key_info: dict[str, Ss58Address]
        ) -> dict[str, int]:

    key2: dict[str, int] = {key_name: balance.get(key_address, 0)
                            for key_name, key_address in local_key_info.items()}

    return key2


def local_keys_to_freebalance(
        c_client: CommuneClient,
        local_keys: dict[str, Ss58Address],
        ) -> dict[str, int]:
    query_all = c_client.query_batch_map(
        {
            "System": [("Account", [])], })
    balance_map = query_all["Account"]

    format_balances: dict[str, int] = {key: value['data']['free']
                                       for key, value in balance_map.items()
                                       if 'data' in value and 'free' in value['data']}

    key2balance: dict[str, int] = concat_to_local_keys(format_balances, local_keys)

    return key2balance


def local_keys_to_stakedbalance(
        c_client: CommuneClient, 
        local_keys: dict[str, Ss58Address],
        netuid: int = 0,
        ) -> dict[str, int]:

    query_all = c_client.query_batch_map(
        {
            "SubspaceModule": [("StakeTo", [netuid])],
        })

    staketo_map = query_all["StakeTo"]

    format_stake: dict[str, int] = {
        key: sum(stake for _, stake in value) for key, value in staketo_map.items()}

    key2stake: dict[str, int] = concat_to_local_keys(format_stake, local_keys)

    return key2stake


def local_keys_allbalance(
        c_client: CommuneClient, 
        local_keys: dict[str, Ss58Address],
        netuid: int | None = None,
    ) -> tuple[dict[str, int], dict[str, int]]:

    staketo_maps: list[Any] = []
    query_result = c_client.query_batch_map(
        {
            "SubspaceModule": [
                ('SubnetNames', [])
            ],
            "System": [("Account", [])],
        })
    balance_map = query_result["Account"]
    all_netuids = list(query_result["SubnetNames"].keys())

    # update for all subnets
    netuids = all_netuids if netuid is None else [netuid]
    for uid in netuids:
        query_result = c_client.query_batch_map(
            {
                "SubspaceModule": [
                    ("StakeTo", [uid]),
                ],
            })
        staketo_map = query_result.get("StakeTo", {})
        staketo_maps.append(staketo_map)

    format_balances: dict[str, int] = {key: value['data']['free']
                                       for key, value in balance_map.items()
                                       if 'data' in value and 'free' in value['data']}

    key2balance: dict[str, int] = concat_to_local_keys(format_balances, local_keys)

    merged_staketo_map: dict[Any, Any] = {}

    # Iterate through each staketo_map in the staketo_maps list
    for staketo_map in staketo_maps:
        # Iterate through key-value pairs in the current staketo_map
        for key, value in staketo_map.items():
            # If the key is not present in the merged dictionary, add it
            if key not in merged_staketo_map:
                merged_staketo_map[key] = value
            else:
                # If the key exists, extend the existing list with the new values
                merged_staketo_map[key].extend(value)

    format_stake: dict[str, int] = {
        key: sum(stake for _, stake in value) for key, value in merged_staketo_map.items()}

    key2stake: dict[str, int] = concat_to_local_keys(format_stake, local_keys)

    key2balance = {k: v for k, v in sorted(
        key2balance.items(), key=lambda item: item[1], reverse=True)}

    key2stake = {k: v for k, v in sorted(
        key2stake.items(), key=lambda item: item[1], reverse=True)}

    return key2balance, key2stake



import typer
from typer import Context
from typing import Optional

from communex._common import BalanceUnit, format_balance
from communex.balance import to_nano
from communex.cli._common import (make_custom_context,
                                  print_table_from_plain_dict)
from communex.compat.key import try_classic_load_key, resolve_key_ss58_encrypted
from communex.errors import ChainTransactionError

balance_app = typer.Typer(no_args_is_help=True)


@balance_app.command()
def show(
    ctx: Context, key: str, 
    unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None
):
    """
    Gets the balances of a key.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)

    with context.progress_status(f"Getting balance of key {key_address}..."):
        subnets = client.query_map_subnet_names()
        netuids = list(subnets.keys())
        balance = client.get_balance(key_address)

        stakes: list[int] = []
        for uid in netuids:
            staketo = client.get_staketo(key_addr=key_address, netuid=uid)
            stakes.append(sum(staketo.values()))

        string_stakes = [format_balance(stake, unit) for stake in stakes]
        netuids = [str(uid) for uid in netuids]
        stake_dict = dict(zip(netuids, string_stakes))

        total = balance + sum(stakes)
        free, total = format_balance(
            balance, unit), format_balance(total, unit)

        print_table_from_plain_dict(
            stake_dict, ["Netuid", "Staked"], context.console)
        print_table_from_plain_dict({"Free": free, "Total": total}, [
                                    "Result", "Amount"], context.console)


@balance_app.command()
def free_balance(
    ctx: Context, key: str, 
    unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None
):
    """
    Gets free balance of a key.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)
    
    with context.progress_status(f"Getting free balance of key {key_address}..."):
        balance = client.get_balance(key_address)

    context.output(format_balance(balance, unit))


@balance_app.command()
def staked_balance(
    ctx: Context, key: str, 
    netuid: int = 0, unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None
):
    """
    Gets the balance staked on the key itself.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)

    with context.progress_status(f"Getting staked balance of key {key_address}..."):
        staketo = client.get_staketo(key_addr=key_address, netuid=netuid)

    result = sum(staketo.values())

    context.output(format_balance(result, unit))


@balance_app.command()
def all_balance(
    ctx: Context, key: str, netuid: int = 0, 
    unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None,
    ):
    """
    Gets entire balance of a key (free balance + staked balance).
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)

    with context.progress_status(f"Getting value of key {key_address}..."):
        staketo = client.get_staketo(key_address, netuid=netuid)
        stake_sum = sum(staketo.values())
        balance = client.get_balance(key_address)
        result = balance + stake_sum

    context.output(format_balance(result, unit))


@balance_app.command()
def get_staked(
    ctx: Context, key: str, netuid: int = 0, 
    unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None,
    ):
    """
    Gets total stake of a key it delegated across other keys.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)

    with context.progress_status(f"Getting stake of {key_address}..."):
        result = client.get_stake(key=key_address, netuid=netuid)

    context.output(format_balance(result, unit))


@balance_app.command()
def transfer(ctx: Context, key: str, amount: float, dest: str):
    """
    Transfer amount to destination using key
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    nano_amount = to_nano(amount)
    resolved_key = try_classic_load_key(key, context)
    resolved_dest = resolve_key_ss58_encrypted(dest, context)

    if not context.confirm(f"Are you sure you want to transfer {amount} tokens to {dest}?"):
        raise typer.Abort()

    with context.progress_status(f"Transferring {amount} tokens to {dest}..."):
        response = client.transfer(
            key=resolved_key, amount=nano_amount, dest=resolved_dest)

    if response.is_success:
        context.info(f"Transferred {amount} tokens to {dest}")
    else:
        raise ChainTransactionError(response.error_message)  # type: ignore


@balance_app.command()
def transfer_stake(ctx: Context, key: str, amount: float, from_key: str, dest: str, netuid: int = 0):
    """
    Transfers stake of key from point A to point B
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    resolved_from = resolve_key_ss58_encrypted(from_key, context)
    resolved_dest = resolve_key_ss58_encrypted(dest, context)
    resolved_key = try_classic_load_key(key, context)
    nano_amount = to_nano(amount)

    with context.progress_status(
        f"Transferring {amount} tokens from {from_key} to {dest} on a subnet with netuid '{netuid}' ..."
    ):
        response = client.transfer_stake(
            key=resolved_key,
            amount=nano_amount,
            from_module_key=resolved_from,
            dest_module_address=resolved_dest,
            netuid=netuid,
        )

    if response.is_success:
        context.info(f"Transferred {amount} tokens from {from_key} to {dest}")
    else:
        raise ChainTransactionError(response.error_message)  # type: ignore


@balance_app.command()
def stake(
    ctx: Context, key: str, amount: float, 
    dest: str, netuid: int = 0,
    ):
    """
    Stake amount to destination using key
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    nano_amount = to_nano(amount)
    resolved_key = try_classic_load_key(key, context)
    resolved_dest = resolve_key_ss58_encrypted(dest, context)

    with context.progress_status(f"Staking {amount} tokens to {dest} on a subnet with netuid '{netuid}'..."):
        response = client.stake(
            key=resolved_key, amount=nano_amount, dest=resolved_dest, netuid=netuid)

    if response.is_success:
        context.info(f"Staked {amount} tokens to {dest}")
    else:
        raise ChainTransactionError(response.error_message)  # type: ignore


@balance_app.command()
def unstake(ctx: Context, key: str, amount: float, dest: str, netuid: int = 0):
    """
    Unstake amount from destination using key
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    nano_amount = to_nano(amount)
    resolved_key = try_classic_load_key(key, context)
    resolved_dest = resolve_key_ss58_encrypted(dest, context)

    with context.progress_status(f"Unstaking {amount} tokens from {dest} on a subnet with netuid '{netuid}'..."):
        response = client.unstake(
            key=resolved_key, amount=nano_amount, dest=resolved_dest, netuid=netuid) # TODO: is it right?

    if response.is_success:
        context.info(f"Unstaked {amount} tokens from {dest}")
    else:
        raise ChainTransactionError(response.error_message)  # type: ignore
